/*************************************************************
Module name: DemandLoadDll.c
Written by:  Jeffrey Richter
Purpose:     Demonstrates how to load a DLL and call 
             functions in it on demand
*************************************************************/


#define STRICT
#include <windows.h>


//////////////////////////////////////////////////////////////


// Here is where you declare the function address variables
// for the functions that you wish to link to on demand.
// Each address is offset from PFN_FIRSTINDEX which points
// to address space that must raise an access violation 
// when we attempt to execute it.  I have chosen an address
// of 0xFFFF0000 which is just below 4GB.
#define PFN_FIRSTINDEX   (0xFFFF0000)


// Dynamically link to MessageBeep
typedef BOOL (WINAPI *PFNMESSAGEBEEP)(UINT);
PFNMESSAGEBEEP g_pfnMessageBeep = 
   (PFNMESSAGEBEEP) (PFN_FIRSTINDEX + 0);


// Dynamically link to MessageBox(A/W)
typedef int (WINAPI *PFNMESSAGEBOX)(HWND, 
   LPCTSTR, LPCTSTR, int);
PFNMESSAGEBOX g_pfnMessageBox = 
   (PFNMESSAGEBOX) (PFN_FIRSTINDEX + 1);


// Make sure that PFN_LASTINDEX is set to the
// last function address index used above.
#define PFN_LASTINDEX   (PFN_FIRSTINDEX + 1)


//////////////////////////////////////////////////////////////


// The DemandLoadDll_ExceptionFilter function changes a 
// thread's program counter. We can restrict the amount
// of CPU-dependent code by defining the PROGCTR macro below.
#if defined(_X86_)
#define PROGCTR(Context)  ((Context).Eip)
#endif

#if defined(_MIPS_)
#define PROGCTR(Context)  ((Context).Fir)
#endif

#if defined(_ALPHA_)
#define PROGCTR(Context)  ((Context).Fir)
#endif

#if defined(_PPC_)
#define PROGCTR(Context)  ((Context).Iar)
#endif

#if !defined(PROGCTR)
#error Module contains CPU-specific code; modify & recompile.
#endif


//////////////////////////////////////////////////////////////


// Macro that places double quotes around the argument
#define Stringize(str)  #str

// Macro for setting dynamic call function information
// dwIndex:    offset from PFN_FIRSTINDEX
// szDllName:  Name of DLL that contains the function
// szFuncName: Name of function contained in the DLL
// pfnProcVar: Name of variable to get the function's address
#define DemandLoadDll_FuncInfo(dwIndex,   \
   szDllName, szFuncName, pfnProcVar)     \
case dwIndex:                             \
   pszDllName   = __TEXT(szDllName);      \
   pszFuncName  = Stringize(szFuncName);  \
   ppfnFuncAddr = (PROC*) &pfnProcVar;    \
   break;

// NOTE: Stringize above expands a function macro to its 
// ANSI/Unicode equivalent.  For example, passing MessageBox
// for szFuncName will cause either MessageBoxA or MessageBoxW
// to be looked-up depending on whether UNICODE is defined
// Of course, you can pass MessageBoxW for szFuncName which 
// will always get the address of MessageBoxW


//////////////////////////////////////////////////////////////


// This exception filter determines if an attempted call 
// is made to a dynamic call function and if so, loads 
// the appropriate DLL, fixes up function address and 
// calls the function.
static LONG DemandLoadDll_ExceptionFilter (
   LPEXCEPTION_POINTERS pep) {

   // Don't swallow exceptions we don't recognize
   LONG lDisposition = EXCEPTION_CONTINUE_SEARCH;

   // Get the address where we attempted the memory access
   DWORD pvAttemptedAddress = 
      pep->ExceptionRecord->ExceptionInformation[1];

   // To resolve a dynamic function call, we need
   // a DLL name, a function name, and a place to 
   // get the resolved function address.
   LPCTSTR pszDllName = NULL;
   LPSTR pszFuncName  = NULL; // Always ANSI; never Unicode
   PROC* ppfnFuncAddr = NULL;

   // We can only handle access violations
   if (EXCEPTION_ACCESS_VIOLATION != 
      pep->ExceptionRecord->ExceptionCode)
      return(lDisposition);

   // If the address is outside the range of our function
   // pointers, we don't know how to handle it
   if ((pvAttemptedAddress < PFN_FIRSTINDEX) || 
       (pvAttemptedAddress > PFN_LASTINDEX))
      return(lDisposition);

   // If there are any problems beyond this point, 
   // execute our exception handler.
   lDisposition = EXCEPTION_EXECUTE_HANDLER;

   // Function addresses are initialized relative to 
   // PFN_FIRSTINDEX so we know which function to resolve
   switch (pvAttemptedAddress - PFN_FIRSTINDEX) {
      DemandLoadDll_FuncInfo(0, "User32", 
         MessageBeep, g_pfnMessageBeep);
      DemandLoadDll_FuncInfo(1, "User32", 
         MessageBox, g_pfnMessageBox);
   }

   // Now that we know how to get the function address, get it
   if (pszDllName != NULL) {

      // If the DLL is not loaded, load it
      HMODULE hinstDll = GetModuleHandle(pszDllName);
      if (hinstDll == NULL) 
         hinstDll = LoadLibrary(pszDllName);

      if (hinstDll != NULL) {

         // Get the address of the function
         *ppfnFuncAddr = GetProcAddress(hinstDll, pszFuncName);

          if (*ppfnFuncAddr != NULL) {

             // We fixed up the address should we ever call 
             // this function again. But for this call, the 
             // thread's PC points to the wrong address; we 
             // need to change PC so that this call succeeds.
             PROGCTR(*pep->ContextRecord) = 
                (DWORD) *ppfnFuncAddr;

             // Continue execution from our new IP address
             lDisposition = EXCEPTION_CONTINUE_EXECUTION;
          }
      }
   }
   return(lDisposition);
}


//////////////////////////////////////////////////////////////


int WINAPI WinMain (HINSTANCE hinstExe, 
   HINSTANCE hinstExePrev, LPSTR szCmdLine, int nCmdShow) {

   // Place an SEH frame at the top of the thread so that
   // dynamically called functions work properly
  __try {

     // Dynamically call MessageBox(A/W)
     g_pfnMessageBox(NULL, "Here is a message box", 
        "DLL loaded on demand", MB_OK);

     // Dynamically call MessageBeep
     g_pfnMessageBeep(MB_OK);

     // Dynamically call MessageBox(A/W) again
     g_pfnMessageBox(NULL, "Here is another message box", 
        "DLL already loaded", MB_OK);
   }
  __except(DemandLoadDll_ExceptionFilter(
     GetExceptionInformation())) {

     // We get in here only if a DLL cannot be 
     // found for a dynamically called function
     MessageBox(NULL, __TEXT("Dynamic function call failed."),
        __TEXT("DemandLoadDll"), MB_OK);
  }

  return(0);
}


///////////////////////// End of File /////////////////////////
